
from csirtg_indicator import Indicator
from csirtg_indicator.constants import COLUMNS


def get_lines(data, cols=COLUMNS, **kwargs):
    if not isinstance(data, list):
        data = [data]

    for i in data:
        if isinstance(i, Indicator):
            i = i.__dict__()

        r = dict()
        for c in cols:
            y = i.get(c, u'')

            if type(y) is list:
                y = u'|'.join(y)

            if c == 'confidence' and y is None:
                y = 0.0

            r[c] = y
            if isinstance(r[c], (str, bytes)):
                r[c] = r[c].replace('\n', r'\\n')

        yield ','.join([str(r[v]) for v in r])





class LookupMixin(object):
    def csirtg(self):
        try:
            from csirtgsdk.search import Search
        except ImportError:
            print('')
            print('The csirtg function requires the csirtgsdk')
            print('$ pip install csirtgsdk')
            print('$ export CSIRTG_TOKEN=1234...')
            print('')
            raise SystemExit

        return Search().search(self.indicator, limit=5)

    def cif(self):
        try:
            from cifsdk.client.http import HTTP as Client
        except ImportError:
            print('')
            print('The cif function requires the cifsdk>=4.0')
            print('$ pip install https://github.com/csirtgadgets/'
                  'verbose-robot-sdk-py/archive/master.zip')
            print('$ export CIF_TOKEN=1234...')
            print('')
            raise SystemExit

        return Client().search({'q': self.indicator, 'limit': 25})

    def farsight(self):
        if self.itype != 'ipv4':
            raise TypeError('%s is not supported' % self.itype)

        try:
            from csirtg_dnsdb.client import Client
        except ImportError:
            print('')
            print('The csirtg function requires the csirtg_dnsdb client')
            print('https://github.com/csirtgadgets/dnsdb-py')
            print('$ pip install csirtg_dnsdb')
            print('$ export FARSIGHT_TOKEN=1234...')
            print('')
            raise SystemExit

        return Client().search(self.indicator)

    def spamhaus(self):
        from csirtg_indicator.utils.spamhaus import ip, fqdn
        if self.is_ip():
            return ip.process(self, resolve_geo=self.resolve_geo)

        elif self.is_fqdn():
            return fqdn.process(self, resolve_geo=self.resolve_geo)

        else:
            return None



FIELDS_CORE = [
    'indicator', 'itype', 'tlp', 'provider', 'group', 'tlp', 'provider',
    'count', 'message', 'tags', 'confidence',
    'description', 'version', 'uuid', 'probability'
]

FIELDS_TIME = [
    'first_at', 'last_at', 'reported_at'
]

FIELDS_META = [
    'application', 'reference', 'reference_tlp', 'data'
]

FIELDS_GEO = [
    'cc', 'latitude', 'timezone', 'longitude', 'city', 'region'
]

FIELDS_IP = [
    'portlist', 'protocol', 'asn', 'asn_desc', 'dest', 'dest_portlist', 'mask',
    'rdata', 'rtype', 'peers'
]

FIELDS_FQDN = [
    'ns', 'mx', 'cname'
]

FIELDS = FIELDS_CORE + FIELDS_GEO + FIELDS_META + FIELDS_IP + FIELDS_TIME \
         + FIELDS_FQDN





class FormatterMixin(object):
    def to_json(self):
        return str(self)

    def to_csv(self):
        from csirtg_indicator.format.csv import get_lines
        return next(get_lines([self]))

    def to_bro(self):
        from csirtg_indicator.format.bro import get_lines
        return next(get_lines([self]))

    def to_snort(self):
        from csirtg_indicator.format.snort import get_lines
        return next(get_lines([self]))

    def to_table(self):
        from csirtg_indicator.format.table import get_lines
        return next(get_lines([self]))

    def to_gexf(self):
        raise NotImplementedError('see: https://networkx.github.io/'
                                  'documentation/networkx-1.10/reference/'
                                  'readwrite.html')




def _load():
    try:
        from csirtg_urlsml_tf import predict as predict_url
        from csirtg_domainsml_tf import predict as predict_fqdn
        from csirtg_ipsml_tf import predict as predict_ip
        from csirtg_ipsml_tf.utils import \
            extract_features as extract_features_ip

    except ImportError:

        print('')
        print(
            'This requires the csirtg_ipsml_tf, csirtg_domainsml_tf and csirtg_'
            'urlsml_tf frameworks')
        print('https://csirtgadgets.com/?tag=machine-learningy')
        print(
            '$ pip install csirtg_ipsml_tf csirtg_domainsml_tf csirtg_urlsml_tf')
        print('')
        raise ImportError


def _to_list(indicators):
    if not isinstance(indicators, list):
        indicators = [indicators]

    return indicators


def _predict_indicators(itype, indicators):
    _load()

    indicators = _to_list(indicators)

    urls = [(i.indicator, idx)
            for idx, i in enumerate(indicators) if i.itype == itype]

    if itype == 'url':
        predict = predict_url([u[0] for u in urls])
    else:
        predict = predict_fqdn([u[0] for u in urls])

    for idx, u in enumerate(urls):
        indicators[u[1]].probability = float(predict[idx][0])

    return indicators


def predict_urls(indicators):
    return _predict_indicators('url', indicators)


def predict_fqdns(indicators):
    return _predict_indicators('fqdn', indicators)


def predict_ips(indicators):
    _load()
    indicators = _to_list(indicators)

    ips = [(i, idx) for idx, i in enumerate(indicators)
           if i.itype == 'ipv4' and not i.probability]

    if len(ips) == 0:
        return indicators

    ips_feats = []
    for i in ips:
        f = list(extract_features_ip(i[0].indicator, i[0].reported_at))
        ips_feats.append(f[0])

    predict = predict_ip([ips_feats])

    for idx, u in enumerate(ips):
        indicators[u[1]].probability = predict[idx][0]

    return indicators



import json
import textwrap
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from datetime import datetime
from base64 import b64encode
import logging
import uuid
import copy
import arrow

from csirtg_indicator.constants import V4_RESERVED, PROTOCOL_VERSION, \
    FIELDS, FIELDS_TIME, LOG_FORMAT, VERSION, GEO, PEERS, FQDN, BASESTRING

from csirtg_indicator.utils import parse_timestamp, resolve_itype, \
    ipv4_normalize

from csirtg_indicator.wrappers.lookups import LookupMixin
from csirtg_indicator.wrappers.predict import PredictMixin
from csirtg_indicator.wrappers.formatter import FormatterMixin
from csirtg_indicator.wrappers.metadata import MetadataMixin
from csirtg_indicator.wrappers.itypes import ItypesMixin
from csirtg_indicator.wrappers.properties import PropertiesMixin


class Indicator(PropertiesMixin, LookupMixin, PredictMixin, FormatterMixin,
                MetadataMixin, ItypesMixin):

    def __init__(self, indicator=None, **kwargs):
        self.version = VERSION

        self._init_fields(**kwargs)

        if indicator:
            self.indicator = indicator

        # geo, fqdn, peers
        self._init_metadata(**kwargs)


    def _init_metadata(self, **kwargs):
        self.resolve_geo = kwargs.get('resolve_geo', GEO)
        self.resolve_peers = kwargs.get('resolve_peers', PEERS)
        self.resolve_fqdn = kwargs.get('resolve_fqdn', FQDN)

        if self.resolve_geo:
            self.geo_resolve()

        if self.resolve_peers:
            self.peers_resolve()

        if self.resolve_fqdn:
            self.fqdn_resolve()

    def _init_fields(self, **kwargs):
        for k in FIELDS:
            # handle these at the end
            if k in ['indicator', 'confidence', 'probability', 'count']:
                setattr(self, f"_{k}", None)
                continue

            if kwargs.get(k) is None:
                v = None

                setattr(self, k, v)
                continue

            # set this at the end
            if k in FIELDS_TIME:
                continue

            if isinstance(kwargs[k], BASESTRING):
                kwargs[k] = kwargs[k].lower()
                if k in ['tags', 'peers']:
                    kwargs[k] = kwargs[k].split(',')

            setattr(self, k, kwargs[k])

        for k in FIELDS_TIME:
            setattr(self, k, kwargs.get(k, None))
        
        self.confidence = kwargs.get('confidence', 0)
        self.count = kwargs.get('count', 1)
        self.probability = kwargs.get('probability', 0)
        self.uuid = kwargs.get('uuid', str(uuid.uuid4()))

    def copy(self, **kwargs):
        try:
            i = Indicator(**copy.deepcopy(self.__dict__()))

            for k in kwargs:
                setattr(i, k, kwargs[k])

            i.uuid = str(uuid.uuid4())
            if not isinstance(i.tags, list):
                i.tags = [i.tags]

            if not kwargs.get('last_at'):
                setattr(i, 'last_at', arrow.utcnow())

        except TypeError:
            i = None

        return i

    def format_keys(self):
        d = self.__dict__()
        for k in d:
            if not isinstance(d[k], str):
                continue

            if '{' not in d[k]:
                continue

            try:
                d[k] = d[k].format(**d)
            except (KeyError, ValueError, IndexError):
                pass

        return Indicator(**d)

    def __dict__(self):
        s = str(self)
        return json.loads(s)

    def __repr__(self):
        i = {}

        for k in FIELDS:

            v = getattr(self, k)
            if not v:
                continue

            if k == 'message':
                v = v.encode('utf-8')

                v = b64encode(v).decode('utf-8')

            if k in FIELDS_TIME and isinstance(v, datetime):
                v = v.strftime("%Y-%m-%dT%H:%M:%S.%fZ")

            if isinstance(v, BASESTRING) and k is not 'message' and \
                    not k.endswith('_at'):
                    v = v.lower()

            if k == 'confidence':
                v = float(v)

            i[k] = v

        sort_keys = False
        indent = None
        if logging.getLogger('').getEffectiveLevel() == logging.DEBUG:
            sort_keys = True
            indent = 4
        try:
            return json.dumps(i, indent=indent, sort_keys=sort_keys,
                              separators=(',', ': '))

        except UnicodeDecodeError as e:
            i['asn_desc'] = unicode(i['asn_desc'].decode('latin-1'))
            return json.dumps(i, indent=indent, sort_keys=sort_keys,
                              separators=(',', ': '))

    def __eq__(self, other):
        d1 = self.__dict__()
        d2 = other.__dict__()
        return d1 == d2


def main():
    p = ArgumentParser(
        description=textwrap.dedent('''\
             Env Variables:
                CSIRTG_INDICATOR_TLP
                CSIRTG_INDICATOR_GROUP

            example usage:
                $ csirtg-indicator -d
            '''),
        formatter_class=RawDescriptionHelpFormatter,
        prog='csirtg-indicator'
    )

    p.add_argument('-d', '--debug', dest='debug', action="store_true")
    p.add_argument('-V', '--version', action='version', version=VERSION)

    p.add_argument('--group', help="specify group")
    p.add_argument('--indicator', '-i', help="specify indicator")
    p.add_argument('--tlp', help='specify tlp', default='green')
    p.add_argument('--tags', help='specify tags')

    args = p.parse_args()

    loglevel = logging.getLevelName('INFO')

    if args.debug:
        loglevel = logging.DEBUG

    console = logging.StreamHandler()
    logging.getLogger('').setLevel(loglevel)
    console.setFormatter(logging.Formatter(LOG_FORMAT))
    logging.getLogger('').addHandler(console)

    i = Indicator(indicator=args.indicator, tlp=args.tlp, tags=args.tags)

    print(i)


if __name__ == '__main__':
    main()



from csirtg_indicator.constants import V4_RESERVED, V6_RESERVED
from csirtg_indicator.utils import is_subdomain
from csirtg_indicator.utils import url_to_fqdn


class ItypesMixin(object):
    itype = None
    indicator = None

    def is_fqdn(self):
        if self.itype == 'fqdn':
            return True

    def is_ip(self):
        if self.itype in ['ipv4', 'ipv6']:
            return True

    def is_ipv4(self):
        if self.itype == 'ipv4':
            return True

    def is_ipv6(self):
        if self.itype == 'ipv6':
            return True

    def is_url(self):
        if self.itype == 'url':
            return True

    def is_hash(self):
        if self.itype in ['md5', 'sha1', 'sha256', 'sha512']:
            return True

    def is_email(self):
        if self.itype == 'email':
            return True

    def is_private(self):
        if not self.itype:
            return False

        if not self.is_ip():
            return False

        if self.is_ipv4() and V4_RESERVED.get(str(self.indicator)):
            return True

        if self.is_ipv6() and V6_RESERVED.get(str(self.indicator)):
            return True

    def is_subdomain(self):
        return is_subdomain(self.indicator)

    def fqdn(self):
        if self.itype == 'fqdn':
            return self.indicator

        if self.itype != 'url':
            return

        return url_to_fqdn(self.indicator)

    def ipv4_to_prefix(self, n=24):
        prefix = self.indicator.split('.')
        prefix = prefix[:3]
        prefix.append('0/%i' % n)
        return '.'.join(prefix)



from csirtg_indicator.utils.geo import process as resolve_geo
from csirtg_indicator.utils.network import resolve_peers
from csirtg_indicator.utils.network import resolve_fqdn, resolve_ns, \
    resolve_url


class MetadataMixin(object):
    indicator = None
    itype = None

    @staticmethod
    def _is_invalid(i):
        if not i:
            return True

        if i in ["", "localhost"]:
            return True

    def _set_rr(self, d, t='A'):
        r = resolve_ns(d, t=t)
        if self._is_invalid(r):
            return

        r = [str(rr) for rr in r]
        setattr(self, t.lower(), r)

    def geo_resolve(self):
        resolve_geo(self)

    def peers_resolve(self):
        resolve_peers(self)

    def fqdn_resolve(self):
        if self.itype not in ['url', 'fqdn']:
            return

        d = self.indicator
        if self.itype == 'url':
            d = resolve_url(self.indicator)
            if not d:
                return

        r = resolve_fqdn(d)
        if self._is_invalid(r):
            return

        if not isinstance(r, list):
            r = [r]

        setattr(self, 'rdata', r)
        setattr(self, 'rtype', 'A')

        for t in ['NS', 'MX', 'CNAME']:
            self._set_rr(d, t)



from prettytable import PrettyTable
import arrow
from csirtg_indicator import Indicator
from csirtg_indicator.constants import COLUMNS, MAX_FIELD_SIZE
from csirtg_indicator.utils import list_to_csv


def _indicator_row(i, cols, max_field_size):
    if isinstance(i, Indicator):
        i = i.__dict__()

    r = []
    for c in cols:
        y = i.get(c, '')

        if isinstance(y, list):
            y = list_to_csv(y)

        if c == 'confidence' and y is None:
            y = 0.0

        if y and (c in ['first_at', 'last_at', 'reported_at']):
            y = arrow.get(y).format('YYYY-MM-DDTHH:mm:ss.SSSSS')
            y = '{}Z'.format(y)
        else:
            y = str(y)
        y = (y[:max_field_size] + '..') if len(y) > max_field_size else y

        r.append(y)

    return r


def get_lines(data, cols=COLUMNS, max_field_size=MAX_FIELD_SIZE):
    t = PrettyTable(cols)

    if not isinstance(data, list):
        data = [data]

    for i in data:

        t.add_row(_indicator_row(i, cols, max_field_size))

    yield str(t)




class PredictMixin(object):
    def predict(self):
        if self.itype not in ['url', 'fqdn', 'ipv4']:
            raise TypeError('%s is not supported' % self.itype)

        try:
            from csirtg_indicator.utils.predict import predict_ips, \
                predict_fqdns, predict_urls

            if self.itype == 'ipv4':
                p = predict_ips(self)
                return [p[0].probability]

            if self.itype == 'fqdn':
                p = predict_fqdns(self)
                return [p[0].probability]

            if self.itype == 'url':
                p = predict_urls(self)
                return [p[0].probability]

        except ImportError:
            print('')
            print('This requires the csirtg_ipsml_tf, csirtg_domainsml_tf and '
                  'csirtg_urlsml_tf frameworks')
            print('https://csirtgadgets.com/?tag=machine-learning')
            print('$ pip install csirtg_ipsml_tf csirtg_domainsml_tf '
                  'csirtg_urlsml_tf')
            print('')
            raise SystemExit



import pytest
from csirtg_indicator import Indicator
from pprint import pprint

DISABLE_TESTS = False
try:
    from csirtg_indicator.utils.predict import predict_urls, predict_fqdns, \
        predict_ips
    from csirtg_urlsml_tf import predict as predict_url

except:
    DISABLE_TESTS = True


@pytest.mark.skipif(DISABLE_TESTS, reason='missing csirtg ml libs..')
@pytest.mark.xfail
def test_indicator_predict():
    i = Indicator('https://g00gle.com/1.html')

    r = predict_urls(i)
    assert r[0].probability > .5

    r = predict_fqdns(Indicator('g00gle.com'))
    assert r[0].probability > .5



import os
import pytricia

from .patterns import *
from .fields import FIELDS, FIELDS_TIME
from .networks import V4_RESERVED, V6_RESERVED

from csirtg_indicator._version import get_versions
__version__ = get_versions()['version']
VERSION = __version__
del get_versions

LOG_FORMAT = '%(asctime)s - %(levelname)s - %(name)s[%(lineno)s] - %(message)s'
PROTOCOL_VERSION = '2.0.0-a.1'

COLUMNS = ['tlp', 'group', 'reported_at', 'indicator', 'asn', 'cc', 'first_at',
           'last_at', 'count', 'tags', 'description', 'confidence', 'rdata',
           'provider', 'probability', 'reference']

MAX_FIELD_SIZE = 30

BASESTRING = (str, bytes)

GEO = os.getenv('CSIRTG_INDICATOR_GEO', False)
if GEO == '1':
    GEO = True

PEERS = os.getenv('CSIRTG_INDICATOR_PEERS', False)
if PEERS == '1':
    PEERS = True

FQDN = os.getenv('CSIRTG_INDICATOR_FQDN', False)
if FQDN == '1':
    FQDN = True



import arrow
from datetime import datetime
import pendulum
import re


def _human_to_dt(ts):
    t = arrow.utcnow()

    if ts == 'now':
        return t

    if ts == 'hour':
        return t.replace(minute=0, second=0, microsecond=0)

    if ts == 'day':
        return t.replace(hour=0, minute=0, second=0, microsecond=0)

    if ts == 'week':
        return t.replace(day=7, hour=0, minute=0, second=0, microsecond=0)

    if ts == 'month':
        return t.replace(day=1, hour=0, minute=0, second=0, microsecond=0)


def _is_valid(ts):
    if isinstance(ts, arrow.Arrow):
        return ts

    t = _human_to_dt(ts)
    if t:
        return t

    try:
        t = pendulum.parse(ts)
    except pendulum.parsing.exceptions.ParserError:
        pass

    t = arrow.get(ts)
    if t:
        return t


def _format_ts(match):
    ts = '{}-{}-{}T{}:{}:{}Z'.format(match.group(1), match.group(2),
                                     match.group(3), match.group(4),
                                     match.group(5), match.group(6))

    t = arrow.get(ts, 'YYYY-MM-DDTHH:mm:ss')
    return t


def _fudge_arrow(ts):
    t = None
    try:
        t = arrow.get(ts)

    except ValueError as e:
        match = re.search(r'^(\d{4})(\d{2})(\d{2})T?(\d{2})(\d{2})(\d{2})Z?$',
                          ts)
        if match:
            return _format_ts(match)

    except arrow.parser.ParserError as e:
        return

    if not t:
        return

    if t.year > 1980:
        return t

    if type(ts) == datetime:
        ts = str(ts)

    if len(ts) == 8:
        ts = '{}T00:00:00Z'.format(ts)
        t = arrow.get(ts, 'YYYYMMDDTHH:mm:ss')

    if t.year < 1970:
        return


def parse_timestamp(ts):
    if isinstance(ts, datetime):
        return arrow.get(ts)

    valid = _is_valid(ts)
    if valid:
        return valid

    valid = _fudge_arrow(ts)
    if valid:
        return valid

    raise TypeError('Invalid Timestamp: %s' % ts)



import arrow
from csirtg_indicator.utils.spamhaus import resolve_dbl as resolve


CODES = {
    '127.0.1.2': {
        'tags': 'suspicious',
        'description': 'spammed domain',
    },
    '127.0.1.3': {
        'tags': 'suspicious',
        'description': 'spammed redirector / url shortener',
    },
    '127.0.1.4': {
        'tags': 'phishing',
        'description': 'phishing domain',
    },
    '127.0.1.5': {
        'tags': 'malware',
        'description': 'malware domain',
    },
    '127.0.1.6': {
        'tags': 'botnet',
        'description': 'Botnet C&C domain',
    },
    '127.0.1.102': {
        'tags': 'suspicious',
        'description': 'abused legit spam',
    },
    '127.0.1.103': {
        'tags': 'suspicious',
        'description': 'abused legit spammed redirector',
    },
    '127.0.1.104': {
        'tags': 'phishing',
        'description': 'abused legit phish',
    },
    '127.0.1.105': {
        'tags': 'malware',
        'description': 'abused legit malware',
    },
    '127.0.1.106': {
        'tags': 'botnet',
        'description': 'abused legit botnet',
    },
    '127.0.1.255': {
        'description': 'BANNED',
    },
}


def process(i, confidence=4, resolve_geo=False):
    if not i.is_fqdn():
        return

    r = resolve(i.indicator)
    r = CODES.get(str(r), None)

    if not r:
        return

    if ' legit ' in r['description']:
        confidence = 1

    i2 = i.copy(**{
        'tags': r['tags'],
        'description': r['description'],
        'confidence': confidence,
        'provider': 'spamhaus.org',
        'reference': 'http://www.spamhaus.org/query/dbl?domain=%s' %
                     i.indicator,
        'reference_tlp': 'white',
        'last_at': arrow.utcnow(),
    })

    if resolve_geo:
        i2.geo_resolve()

    return i2


def main():
    from csirtg_indicator import Indicator
    i = Indicator('ns2.ndxylfpxuwowlhycfh.pw')
    import logging
    logger = logging.getLogger('')
    logger.setLevel(logging.DEBUG)

    print(process(i))


if __name__ == '__main__':
    main()
